/*******************************************************************************
 $Author: dmc $
 $Revision: 3223 $
 $Date: 2023-11-16 18:47:40 -0400 (Thu, 16 Nov 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/gsdp-server.js $
*******************************************************************************/

// set the log level for the log4js library
var log4js = require( "log4js" );
var os = require( "os" );


var logAppenders = [
    {
        "type": "logLevelFilter",
        "level": "WARN",
        "appender": {
                type: "console"
            }
    }



];

log4js.configure( {
    appenders: logAppenders
} );

// default global loglevel is set in the appender with no 'category'
// Don't use these:
//log4js.setGlobalLogLevel( "ERROR" );
//log4js.setGlobalLogLevel( "DEBUG" );
//log4js.setGlobalLogLevel( "TRACE" );

var net = require( "net" );

var log = log4js.getLogger( "gsdp-server.js" );

var shots = [],
    shotCounter = 0,
    lastServerPort = 11001;

log.info( "gsdp-server starting..." );

// Possible TODO:
// conn.on( "get-shot", function( data ) {
//     var index = parseInt( data.data.number, 10 );
//     if ( shots[ index ] !== undefined ) {
//         conn.send( data.socket, "get-shot", { result: true, shot: shots[ index ], number: index } );
//     } else {
//         conn.send( data.socket, "get-shot", { result: false, shot: undefined, number: index  } );
//     }
// } );


var clients = [], gsdpServerPort = 1967;

global.gsdp_clients = clients;

net.createServer( function( socket ) {
    socket.name = socket.remoteAddress + ":" + socket.remotePort;
    var messagehandlerJSON = require( "./lib/messagehandlerJSON/messagehandlerJSON.js" );

    gsdp_server_log("gsdp-server new connection; socket.setNoDelay(), clients.length=",clients.length);
    socket.setNoDelay(); // so socket.write() is delivered immediately

    var client = undefined;

    // search existing socket... reuse client if possible
    for ( var i = 0; i < clients.length; i++ ) {
        if ( clients[ i ].ip === socket.remoteAddress) {
            if (!client) client = clients[i]; //re-use the first one we find
            if (clients[i].active) {
                clients[ i ].active = false;
                logStatistic( clients[ i ], "LP" );
                logStatistic( clients[ i ], "D" );
                clients[ i ].socket.destroy(); // clean up any other incoming connections from same IP
            }
        }
    }

    if (!client) {
        client = {}; // new
        clients.push( client );
    } else {
        gsdp_server_log("gsdp-server re-use client, client.most_recent_shot_no=",client.most_recent_shot_no);
    }
    client.socket = socket;
    client.ip = socket.remoteAddress;
    client.port = socket.remotePort;
    client.last = Date.now();
    client.start = Date.now();
    client.heartbeats = 0;
    client.active = true;
    client.firstHeartbeat = undefined;
    client.lastHeartbeat = undefined;
    client.localIp = socket.address().address;
    if (client.itimer) clearInterval(client.itimer);
    client.itimer = undefined;
    /* fields which will be filled in by orf_updater
    remote_rf_level, temperature, serialno
    connected, ORF->true, ORF-TIMEOUT -> false
    */
    // client.circBufSize = 48; // will be set by shot-fa
    client.circBufSize = undefined; // will be set by shot-fa
    client.fn_request_shots = request_shots;
    client.fn_refetch_needed = refetch_needed;
    client.fn_do_refetch = do_refetch;
    client.get_shot_number_mod = get_shot_number_mod;
    gsdp_server_log("gsdp-server client.ip/port=",client.ip,client.port);

    function connect_client_to_global() { // connect "client"
        client.device = undefined; // first, disconnect... in case it isn't found
        for ( var i in global.data.devices ) {
            var d = global.data.devices[ i ];
            if ( d.type === "GSDP" && d.serialno === client.serialno ) {
                client.device = d;
                exists = true;
                log.trace('orf_updater, FOUND device');
                break;
            }
        }
    }

    // logStatistic( client, "C" );

    // handler for incoming shots
    var cmsgHandler = new messagehandlerJSON.cmsgHandler();

    cmsgHandler.on("hello", function (jsondata) {
        console.log("cmsg handle hello",jsondata)
        client.serialno = jsondata.id;
        client.connected = true;
        emitBroadcast(client); // will add my "client" to global.data.devices if it isn't already there
        console.log("cmsg global.data.devices=",global.data.devices)
        connect_client_to_global();
        socket.write(JSON.stringify(
            {"srsp" : "OK"}
        ) + "\n");
    });

    cmsgHandler.on("ping", function (jsondata) {
        console.log("cmsg handle ping",jsondata)
        client.nxtmeas = jsondata.nxtmeas; // required field
        socket.write(JSON.stringify(
            {"srsp" : "OK"}
        ) + "\n");

    });

    cmsgHandler.on("meas", function (jsondata) {
        console.log("cmsg handle meas",jsondata)
        socket.write(JSON.stringify(
            {"srsp" : "OK"}
        ) + "\n");
    });

    cmsgHandler.on("new-shot", function (shotModel) {
        console.log("handling cmsgHandler new-shot, immediately exiting");
        return;
        client.circBufSize = shotModel.circBufSize;
        gsdp_server_log(client.circBufSize,'-',"gsdp-server new_shot", shotModel.shotNumber, "client=", client.serialno, shotModel.shotCountVector, client.next_shot_no);
        // bugfix1 - before first ORF is sent, client.serialno = client.device = undefined; therefore shots
        // handed to "measFromBeam" will be discarded
        if (!client.device) {
            gsdp_server_log("BUGFIX1, new-shot: most_recent=",client.most_recent_shot_no,'MR-1=',client.get_shot_number_mod(shotModel.shotCountVector-1),'!client.device=',!client.device,'client.device=',client.device);
            if (client.most_recent_shot_no == undefined) {
            	client.most_recent_shot_no = client.get_shot_number_mod(shotModel.shotCountVector-1);
            }
            //gsdp_server_log('shotModel=',shotModel);
            gsdp_server_log("BUGFIX1... ignoring for now; refetch will happen later");
            return;
        }
        if (client.fn_refetch_needed(shotModel.shotCountVector)) {
            gsdp_server_log("new-shot: refetch needed!");
            if (global.data && global.data.config && global.data.config.solo_test_no_refetch_new_shot) {
                gsdp_server_log("... but, ignoring");
                ; // ignore error, carry on
            } else {
                gsdp_server_log("... discard & refetch");
                client.fn_do_refetch(); // ask for shots to be re-sent, ignore this shot for now, we
                return; // will later request it to be resent and we'll deal with it then.
            }
        }

        var shot = {
            shotnum: shotModel.shotNumber, // 0..65535
            counters: [
                shotModel.counter0,
                shotModel.counter1,
                shotModel.counter2,
                shotModel.counter3,
                shotModel.counter4,
                shotModel.counter5,
                shotModel.counter6,
                shotModel.counter7
            ],
            raw_counters: shotModel.raw_counters,
            G2_temp_rdg: shotModel.mean_temp_uncal,
            out_set: 0,
            GPIO18: shotModel.GPIO18, // JRPL target position reading
            mcucount: shotModel.shotCountVector // 0..46(e.g.)
        };
        if (global.data.config.global_temperature !== undefined) {
            shot.G2_temp_rdg = global.data.config.global_temperature;
        }

        var logshot = {
          ts: Date.now(),
          source: "SOLO",
          shot: shot
        };

        log.trace( "LOGSHOT = ", logshot );
        if (global.data && global.data.config && global.data.config.solo_test_discard_shots > 0) {
			global.data.config.solo_test_discard_shots--;
			gsdp_server_log("gsdp-server solo_test_discard_shots #",shotModel.shotCountVector
			  ,'mr_shot=',client.most_recent_shot_no, 'next_shot_no=',client.next_shot_no
			  ,'discards remaining:',global.data.config.solo_test_discard_shots
			);
			; // ignore shot, do not process
        } else {
            // When we receive a shot, the previously-reported next_shot_no might be out of date,
            // so clear it; it will be correctly set by the next ORF.
            client.most_recent_shot_no = shotModel.shotCountVector;
            client.next_shot_no = undefined; // wait until the next ORF
            gsdp_server_log("DMC gsdp-server process shot");
            global.qndEventEmitter.emit( "measFromBeam", {
                beamMeasurement: logshot,
                srcDevice: client.device
            } );
        }
    } );

    var errHandler = new messagehandlerJSON.errHandler();
    errHandler.on("error", function (parse_error) {
        socket.write(JSON.stringify(
            {"srsp" : "ERROR", errors: parse_error}
        ) + "\n");
    })

    var proxy = new messagehandlerJSON.Proxy(); // hardcoded in there is the knowledge that 0xBB is our framing byte
    proxy.addHandler( "cmsg", cmsgHandler );
    proxy.addHandler( "ERROR", errHandler );
    // proxy.addHandler( "ping", shotgsdp );
    // proxy.addHandler( "meas", shotgsdp );


    // Handle incoming messages from clients.
    socket.on("data", function (data /* typeof == Buffer */) {

        client.last = Date.now();
        var data_str = data.toString( "ascii" ); // 'ascii' gives byte-for-byte conversion
        // var buf = new Buffer( data );

        // log.trace( "RAW DATA IN " + socket.remoteAddress + ":" + socket.remotePort + ": " + buf.toString( "hex" ) );
        //log.trace( "RAW DATA str= ", str )
        if (!client.active) {
            return;
        }

        proxy.addData(data_str);

        do_debug_checks(client);

    });

    socket.on( "error", function( err ) {
    	gsdp_server_log( 'socket.on(error,',socket.name,'): ' + err );
        logStatistic( client, "E" );
    } );

    // Remove the client from the list when it leaves
    socket.on( "end", function( ) {
    	gsdp_server_log( "client disconnected (" + socket.name + ")" );
        client.active = false;
        client.connected = false;
        logStatistic( client, "LP" );
        logStatistic( client, "D" );
        emitBroadcast( client ); // update the fact that we are offline
    } );

} ).listen( gsdpServerPort );

log.info( "gsdp-server started listening on port " + gsdpServerPort );

function do_debug_checks(client) {
    if (!data || 'object' != typeof data.config) {
        gsdp_server_log("DMC do_debug_checks() missing global.data")
        return;
    }

    if (data.config.solo_test_req_shots) {
        var i = data.config.solo_test_req_shots.shot_no;
        var j = data.config.solo_test_req_shots.n_shots || 1;
        client.fn_request_shots(i,j);
        // var cmd = new Buffer("send shot:") + new Buffer([i, j]);
        // console.log("Request shot#", i, "nshots=", i, "cmd=", Buffer(cmd.toString("hex")));
        // client.socket.write(cmd);
        apply_crud({
            config: {
                solo_test_req_shots: 0 // null, undefined don't work
                ,dmc_was_here:1
            }
        });
        gsdp_server_log("DMC test_req_shots config=",data.config.solo_test_req_shots,data.config.dmc_was_here);
    }

}

// 'client' member function
function request_shots(shot_nr, n_shots) {
	if (!this.socket || !this.socket.writable) {
	    log.trace("request_shots -- socket NOT WRITABLE",this.socket.name,this.socket.writable);
		return;
	}
    var cmd = new Buffer("send shot:") + new Buffer([shot_nr, n_shots]);
    this.socket.write(cmd);
    gsdp_server_log("request shots #", shot_nr, "nshots=", n_shots, "cmd=", Buffer(cmd.toString("hex")));
}

// 'client' member function
function get_shot_number_mod(n) { // mod-safe
    return ((n % this.circBufSize) + this.circBufSize) % this.circBufSize;
}

// 'client' member function
function refetch_needed(incoming_shot_no) { // boolean
    // ORF sets my_client.next_shot_no
    // successfully received shot sets client.most_recent_shot_no, which will be undefined until first shot reading arrives
    //
    log.trace('check refetch_needed(',this.serialno,')...... this.most_recent_shot_no',this.most_recent_shot_no,'this.next_shot_no',this.next_shot_no);
    if (typeof this.most_recent_shot_no != 'number') return false;
    if (typeof this.next_shot_no != 'number') return false;
    if (isNaN(this.next_shot_no)) return false;
    
    if (this.circBufSize == undefined) { // then we have not received a shot since client connnection
    	// was made, and we are being called in order to process an incoming ORF.  Note that an ORF
    	// will have set this.next_shot_no
    	// Since a shot is our ONLY source of the circular buffer size, we therefore
    	// can't do sane/rational buffer-wrap computations
    	return true; // do_refetch will have special code for case this.circBufSize == undefined && !isNaN(this.next_shot_no)
    }
    
    if (typeof incoming_shot_no == 'number') { // decide if incoming shot# is the one we are waiting for?
        if (this.get_shot_number_mod(this.most_recent_shot_no+1-incoming_shot_no) == 0) {
            log.trace('refetch_needed()...... FALSE 1...incoming_shot_no is the one we were waiting for');
            return false;
        } {
            log.trace('refetch_needed()...... TRUE 1...incoming_shot_no '
            ,incoming_shot_no,"*NOT* the one we want",this.most_recent_shot_no+1);
            return true;
        }
    }

	// incoming_shot_no is not a number ==> this happens when we are called by orf_updater()
    if (this.get_shot_number_mod(this.next_shot_no-1) != this.get_shot_number_mod(this.most_recent_shot_no)) {
        log.trace("refetch_needed() TRUE 2",this.next_shot_no,this.most_recent_shot_no,
        'modcheck:',this.get_shot_number_mod(this.next_shot_no-1),this.get_shot_number_mod(this.most_recent_shot_no));
        return true; // mismatch; need to refetch
    }
    log.trace('refetch_needed()...... FALSE 2');
    return false;
}

// 'client' member function
function do_refetch() {
    if (this.circBufSize == undefined) { // then we have not received a shot since client connnection
    	// was made, and we therefore have NO idea about the size of the circular buffer, and therefore
    	// can't do sane/rational computations on it
    	// so as to discover this.circBufSize, we must ask for ANY shot at all, except for shot # = most_recent_shot_no+1
    	if (!isNaN(this.most_recent_shot_no)) {
	        this.fn_request_shots(this.most_recent_shot_no, 1);
    	} else if (undefined != this.next_shot_no ) { // if server has been restartedt hen most_recent_shot is unknown, in which case, wait until
    	} else {
    		// eventually we will get an ORF, which will set this.next_shot_no
    	}
    	return;
	}


    var n_shots_needed = this.get_shot_number_mod(
        this.next_shot_no - 1 - this.most_recent_shot_no
    );
    var max_shots_to_fetch = 15; // as many as 20
    if (n_shots_needed > max_shots_to_fetch) n_shots_needed = max_shots_to_fetch;
    var next_shot_to_get = this.get_shot_number_mod(this.most_recent_shot_no + 1);
    gsdp_server_log("do_refetch(), n_shots_needed =",n_shots_needed,"next_shot_to_get =", next_shot_to_get
      ,'('+this.circBufSize+')'
      , 'next_shot_no=',this.next_shot_no,'most_recent_shot_no=',this.most_recent_shot_no
    );
    if (n_shots_needed) {
        gsdp_server_log("do_refetch recent,next,num:", this.most_recent_shot_no, this.next_shot_no, "==>", n_shots_needed,
            "next_shot_to_get, n_shots_needed==>",next_shot_to_get, n_shots_needed);
        this.fn_request_shots(next_shot_to_get, n_shots_needed);
    }
}


setInterval( function() {
    var logStat = log4js.getLogger( "gsdp-server-stats.js" );
    logStat.debug( "______________________________________________________________________________" );
    logStat.debug( "Statistics:" );

	var prevClientAssoc = {}; // to calculate time lost between successive connections
	var nActive = 0;
    for ( var i = 0; i < clients.length; i++ ) {
        var c = clients[ i ];
		var prevClient = prevClientAssoc[ c.ip ];
        var msg = "";

        nActive += ( !!c.active );

        msg += ( !c.active ? "X..." : "" );
        msg += "IP: " + c.ip + ":" + c.port;
        msg += "; heartbeats: " + c.heartbeats;
        msg += "; start: " + new Date( c.start ).toLocaleTimeString();
        msg += "; active: " + Math.round( ( c.last - c.start ) / 1000 ) + "s";
        msg += "; last package: " + Math.round( ( Date.now() - c.last ) / 1000 ) + "s";
        if ( prevClient ) {
            msg += "; coverage gap: " + Math.round( ( c.start - prevClient.last ) / 1000 ) + "s";
        }
        logStat.debug( msg );
		prevClientAssoc[ c.ip ] = c;
    }
    logStat.debug( "number of created sockets: " + clients.length, "Active clients: ", nActive );

}, 10000 );

setInterval( function() {
    var now = Date.now() - 7000;
    for ( var i = 0; i < clients.length; i++ ) {
        var c = clients[ i ];
        if ( c.active && ( c.lastHeartbeat === undefined || c.lastHeartbeat < now ) && !c.invalid ) {
            logStatistic( c, "LP" );
            logStatistic( c, "C" );
            // DMC: validity timeout?
            // c.invalid = true;
        }
    }
}, 7000 );

var emitBroadcast = function (client) {
    var device_msg = {
        type: "GSDP",
        serialno: client.serialno, // set by cmsg="hello"
        // hwversion: client.hwversion, fwversion: client.fwversion, // set by huhu-ORF
        // remote_rf_level: client.remote_rf_level,
        // temperature_uncal: (typeof client.mean_temp_uncal == 'number') ?
        //     (+client.mean_temp_uncal.toFixed(1)) : client.mean_temp_uncal, // uncalibrated SOLO temperature_reading 
        connected: !!client.connected, // RED/GREEN state? (late ORF?)
        name: "",
        idx: client.idx,
        deviceIP: client.ip
    };
    global.qndEventEmitter.emit( "deviceBroadcastRecv", { message: device_msg, ip: client.localIp } );
};

setInterval( function() {
    console.log("\n\n\n\ngsdp=server loop");

    for ( var i = 0; i < clients.length; i++ ) {
        var c = clients[ i ];
        console.log("client #",i,"active=",c.active,"invalid=",c.invalid,"device=",c.device)
        if ( c.active && !c.invalid && c.device) {
        	c.idx = i;
            emitBroadcast( c );
        }
    }

}, 3000 );

function gsdp_server_log() {
    if (global.data && global.data.config && global.data.config.gsdp_server_log) { // controlled by boolean data.config.gsdp_server_log
        log.debug.apply(log,Array.prototype.slice.call(arguments));
    }
	// console.log.apply(log,Array.prototype.slice.call(arguments));
}

var logStatistic = function( client, type ) {
    var logger = log4js.getLogger( "gsdp-server-uptimes.js" );
    logger.setLevel("INFO");
    var msg = Date.now();

    if ( type === "LP" && client.lastHeartbeat !== undefined ) {
        msg = client.lastHeartbeat;
    } else if ( type === "LP" ) {
        return;
    }

    msg += "," + client.ip;
    msg += "," + type;
    logger.info( msg );
}

exports.initialize = function(data) { // messy..
    global.data = data;
}