/*******************************************************************************
 viewer_targetrow.js: client module for target row display

 $Author: dmc $
 $Revision: 3206 $
 $Date: 2023-07-06 10:53:57 -0300 (Thu, 06 Jul 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_targetrow.js $
*******************************************************************************/

targetrow = new function() {
	
	this.show_targets = false; // for toggling display of target images

	this.init = function() {
		this.div = $("#divTargetRow");

	}

	this.clear = function() {
		hide(this.div);
		this.div.empty();
        this._isInit = false;
	}

	this.select_string = function(id) {
				
		// save cookie for this frame in list
		set_cookie("recentframe", g_strings[id].frame().id());
		
		select_string(id);
		
        this.redraw();
	}

    this.selectFrame = function(id) {
        var stringId = g_frames[id].string().id();
        this.select_string(stringId);
    }
    
    /**
     * Attention: this function called every second (on every heartbeat)
     */
    this._isInit = false;
	this.update = function() {
		if (!live_mode()) {
            // do not display when viewing old strings
            this.clear();
            return; 
        } else if (!this._isInit) {
            this.redraw();
            this._isInit = true;
        } else {
            this._checkForcedToSpecificFrame();            
        }
	}

    this.redraw = function() {
        this.clear();
        this._init();
    }
    
    this.updateByString = function(data) {
        if (!this.show_targets) return;
        
        var frameId = data.frame().id(),
            $item = this._getItemByFrameId(frameId),
            $div = $item.find('div.thumbnail:first'),
            curShot = data.cur_shot(),
            currentShotNumber = $item.data('current-shot-number');
        if (
            (curShot === undefined && currentShotNumber === undefined) || 
            (currentShotNumber !== undefined && curShot !== undefined && currentShotNumber === curShot.num)
        ) {
            // nothing changed, so ignore new rendering...
            return;
        }
        
        $div.find('img').remove();
        this.draw_small_target($div, data);
        if (curShot !== undefined) {
            $item.data('current-shot-number', curShot.num);
        } else {
            $item.data('current-shot-number', undefined);
        }
    }
    
    this._getItemByFrameId = function(id) {
        return $('[data-targetrow-frame-id=' + id + ']');
    }
    
    this._checkForcedToSpecificFrame = function() {
		// check if we should be forced to a specific frame
		// warning: infinite loop hazard... make sure select_string works
		for (var i in g_frames) {
			var frame = g_frames[i];
			var str_id = frame.string().id();
			if (frame_is_valid_to_select(frame) && (user_is_shooter(str_id) >= 0 || user_is_scorer(str_id))) {
				if (cur_string().id() != str_id) {
					log("currently on " + cur_string().frame().data.name + ", forcing to " + frame.data.name);
					this.select_string(str_id);
					update(); // attention global function (@see: viewer.js)
					return;
				}
				break;
			}
		}
    }
    
    this._init = function() {
		var table = $("<table>"),
            tr = $('<tr>').appendTo(table);
		this.div.append(table);

		// enable scrolling when scrolling is possible
		// still it is possible to scroll vertically - not sure how to catch this
		table.bind("touchstart", function(e) {
			e.originalEvent.allow_touchstart = true;
		});

		table.bind("touchmove", function(e) {
			
			// comment this line to allow scrolling (not sure why...)
			//if (divTargetRow[0].scrollWidth <= divTargetRow[0].clientWidth) return;
			// in either case a possibility of scrolling the page when dragging the target row is acceptable
			
			e.originalEvent.allow_touchmove = true;
		});
		
		var selected_td = null;
		
		// determine frames to display in target row, and sort for display
		var frames_to_display = [];
		for (var i in g_frames) {
			var frame = g_frames[i];
			if (!frame_is_valid_to_select(frame)) continue; // check against admin/frame status
			frames_to_display.push(i);
		}
		
		frames_to_display.sort(function(a,b) {
			if (g_frames[a].data.name == g_frames[b].data.name) return 0;
			if (g_frames[a].data.name == "new") return -1;
			if (g_frames[b].data.name == "new") return 1;
			return g_frames[a].data.name.toUpperCase().localeCompare(g_frames[b].data.name.toUpperCase());
		});
		
		// special "more" button if needed
		if (display_admin_mode(true) && !toolbar_system.show_hidden_frames && frames_to_display.length != numberof(g_frames)) {

			var td = $("<td>");

			var number = numberof(g_frames) - frames_to_display.length;
			var item = toolbar.new_button(td, S("FRAME_HIDDEN", number), function() {
				toolbar_system.show_hidden_frames = true;
				update();
			});
			item.css("font-size", "16px");
			item.css("font-weight", "bold");
			item.css("width", "70px");
			item.css("height", "50px");
			item.css("padding", "0px");
			item.css("margin", "0px");

			td.css("vertical-align", "middle");

			tr.append(td);
		}

		for (var i in frames_to_display) {
			
			var frame = g_frames[frames_to_display[i]];

			// select frame on click
			var td = $("<td>");
			if (frame.string() == cur_string()) {
				td.mousedown(function() {
					targetrow.show_targets = !targetrow.show_targets;
					if (targetrow.show_targets) dequad_refresh_targetrow(); // ask for shots to be updated soonish
                    targetrow.redraw();
					//update();
				});
			} else if (user_is_shooter() < 0 && !user_is_scorer()) { // allow switching
				td = $("<td onmousedown=\"targetrow.selectFrame('" + frame.id() + "'); update();\">");
			}
            td.attr('data-targetrow-frame-id', frame.id());

			// special styling for selected frame
			if (frame.string() == cur_string()) {
				td.addClass("selected");
				selected_td = td;
			}

			var frame_title = $("<div>").addClass('label').text(frame.data.name);
			frame_title.css("color", "black");
			frame_title.css("white-space", "nowrap");

			if (!frame.data.enabled) frame_title.css("background-color", "#999");
			else if (!frame.device()) frame_title.css("background-color", "#FF0");
			else if (!frame.data.connected || !g_connected) frame_title.css("background-color", "#F88");
			else frame_title.css("background-color", "#8F8");
			td.append(frame_title);
			
			//frame_title.css("background-color", "#8F8"); // HARD CODED
			
			if (frame.data.enabled && display_admin_mode(true)) {
				
				var assoc_text = S("FRAME_UNASSOCIATED");
				if (frame.device()) assoc_text = frame.device().device_type_and_short_name();
				
				frame_assoc = $("<div>").css("width", "100%").text(assoc_text);
				frame_assoc.css("color", "black");
				frame_assoc.css("white-space", "nowrap");
				frame_assoc.css("background-color", "#FF0");
				td.append(frame_assoc);
			}

			// canvas size fills this area
			td.addClass("targetrow-item");
			tr.append(td);
            
			if (this.show_targets) { // allowing in admin mode now, because scroll bug is fixed
				var $thumbnail = $('<div />').addClass('thumbnail');
                td.append($thumbnail);
				this.draw_small_target($thumbnail, frame.string());
			}

		}
		
		// G2 only: show all devices, providing links to the other targets
		if (g_config.device_type == "G2") {
			for (var i in g_devices) {	
				var device = g_devices[i];
				
				// don't show if it's not a target or it's already associated to a frame
				if (!device.data.is_target || device.frame()) continue;
				
				if (device.data.is_present == false) continue; // only if is_present or is not set (G1)
				
				// do not show any with allow_association (self)
				if (device.data.allow_association) continue;
				
				var td = $("<td>");
	
				var frame_title = $("<div>").css("width", "100%");
				
				// display IP or name of associated frame
				if (display_admin_mode(true)) {
					if (device.data.name) frame_title.text(device.data.name);
					else frame_title.text(device.data.type);
				} else {
					if (device.data.name) frame_title.text(device.data.name);
					else frame_title.text(device.device_type_and_short_name());
				}
				
				frame_title.css("color", "black");
				frame_title.css("white-space", "nowrap");
				frame_title.css("background-color", "#88F");
				td.append(frame_title);
				
				
				if (display_admin_mode(true)) {
					frame_assoc = $("<div>").css("width", "100%").text(device.device_type_and_short_name());
					frame_assoc.css("color", "black");
					frame_assoc.css("white-space", "nowrap");
					frame_assoc.css("background-color", "#FF0");
					td.append(frame_assoc);
				}
				
				// canvas size fills this area
				td.css("min-width", "80px"); // needed to work in safari
				td.css("width", "80px");
	
				if (this.show_targets) {
					td.css("height", "80px");
					this.draw_device_target(td, device);
				}
	
				if (!display_admin_mode(true)) { // for G2's to select neighboring targets
					td.wrapInner($('<a style="text-decoration: none" href="http://' + device.data.ip + '"/>'));
				}
	
				tr.append(td);
			}
		}

		// need to compute div width manually
		var width = table.width();
		var width_parent = this.div[0].parentElement.clientWidth;
		if (width > width_parent - 20) width = width_parent - 20;
		this.div.css("width", width + "px");

		// scroll to center selected frame
		if (selected_td != null) {
			var center_div = this.div[0].clientWidth / 2;
			var center_face = selected_td[0].offsetLeft + selected_td[0].clientWidth / 2;
			this.div.scrollLeft(center_face - center_div);
		}

		//log("targetrow.update() done");
		show(this.div);
    }
    
	this.draw_small_target = function(div, string) {
		
		var args = {}; // for passing parameters around
		args.string = string;
		args.frame = string.frame();
		args.face = string.face();
		
		var half_mast = !args.frame.data.connected || !args.frame.data.enabled || !g_connected;
		half_mast = false; // always; DMC 3-May-2023

		// zoom
		var zoom_small = maintarget.zoom * 0.75; // a little less than normal
		if (half_mast) zoom_small = 1.1 * args.face.data.targetboard.width * args.face.mm_to_moa; // disconnected


		// set up display area
		args.w = div.width();
		args.h = div.height();
		args.cx = args.w / 2;
		args.cy = args.h * ((!half_mast)?(0.5):(1.0)); // position down if not connected
		args.moa_to_px = Math.min(args.w, args.h) / zoom_small;
		args.mm_to_px = args.face.mm_to_moa * args.moa_to_px;

		// try and keep canvas with the frame object
		if (!ok(args.frame.targetrow_canvas)) {
			args.frame.targetrow_canvas = $("<canvas>").attr({width:args.w, height:args.h});
		}

		var g = args.frame.targetrow_canvas[0].getContext("2d");

		g.clearRect(0, 0, args.w, args.h);

		var fclass = draw_fclass_target();
		args.face.draw_target_thumbnail(g, args, fclass);

        var shot = args.string.cur_shot();
        if (shot && shot.v) {
            if (isShotHiddenByDelay(shot, args.string)) {
                // shot is hidden by delay, update the target again after n seconds...
                // TODO: already started timeout should be stopped before starting a new one!
				var now = (new Date()).getTime() - cur_ts_local_error,
                    timeRemaining = shot.ts + g_config.shot_display_delay * 1000 - now,
                    self = this;
                setTimeout(function() { 
                    div.find('img').remove();
                    self.draw_small_target(div, string); 
                }, timeRemaining);
            } else {
                args.face.draw_only_selected_shot(g, args, fclass);
            }
        }
		
		// add canvas to page
		//div.append(args.frame.targetrow_canvas);

		// convert canvas to normal image and add that instead
		// to fix draw issues
		var image = new Image();
		image.src = args.frame.targetrow_canvas[0].toDataURL("image/png");
		div.append(image);

	}

	this.draw_device_target = function(div, device) {
		
		var args = {}; // for passing parameters around
		args.string = cur_string();
		args.frame = cur_string().frame();
		args.face = cur_string().face();
		args.device = device;
		
		var half_mast = !args.frame.data.connected || !args.frame.data.enabled || !g_connected;

		// zoom
		var zoom_small = maintarget.zoom * 0.75; // a little less than normal
		if (half_mast) zoom_small = 1.1 * args.face.data.targetboard.width * args.face.mm_to_moa; // disconnected

		// set up display area
		args.w = div.width() - 5;
		args.h = div.height() - 5;
		args.cx = args.w / 2;
		args.cy = args.h * ((!half_mast)?(0.5):(1.0)); // position down if not connected
		args.moa_to_px = Math.min(args.w, args.h) / zoom_small;
		args.mm_to_px = args.face.mm_to_moa * args.moa_to_px;

		// try and keep canvas with the frame object
		if (!ok(args.device.targetrow_canvas)) {
			args.device.targetrow_canvas = $("<canvas>").attr({width:args.w, height:args.h});
		}

		var g = args.device.targetrow_canvas[0].getContext("2d");

		g.clearRect(0, 0, args.w, args.h);

		args.face.draw_target_thumbnail(g, args, draw_fclass_target());
		args.face.draw_device_recent_shot(g, args);

		// add canvas to page
		div.append(args.device.targetrow_canvas);
		
	}

}

