/*******************************************************************************
 $Author: dmc $
 $Revision: 3170 $
 $Date: 2023-02-21 21:20:26 -0400 (Tue, 21 Feb 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_toolbar_system.js $
*******************************************************************************/

// bugfix: to avoid "losing" focus on file-selection, don't be continuously destroying/redrawing it..
// .. use this global <input type='file'> object
var glob_input = $("<input type='file' name='selected_file'>"); // multiple?
glob_input.css("padding-left", "10px");

var toolbar_system = new function() {

	this.simulate_mode = false; // shot simulator
	this.visualcal_mode = false; // allow hold and drag to calibrate
	this.show_hidden_frames = false; // show hidden frames on target row
	this.selected_user = 0; // user selected in tab for editing
	
	this.event_list = null;
	
	this.init = function() {
		this.icon = $("<img src='images/admin.png' class='toolbar_icon'>");
		this.tab_icon = [];
		this.tab_icon["display"] = $("<img src='images/display.png' class='toolbar_icon'>");
		this.tab_icon["frame"] = $("<img src='images/admin_frame.png' class='toolbar_icon'>");
		this.tab_icon["faces"] = $("<img src='images/admin_face.png' class='toolbar_icon'>");
		this.tab_icon["users"] = $("<img src='images/user.png' class='toolbar_icon'>");
		this.tab_icon["storage"] = $("<img src='images/storage.png' class='toolbar_icon'>");
		this.tab_icon["rstat"] = $("<img src='images/database.png' class='toolbar_icon'>");
		this.tab_icon["advanced"] = $("<img src='images/advanced.png' class='toolbar_icon'>");
		this.tab_icon["gpios"] = $("<img src='images/gpios.png' class='toolbar_icon'>");
		this.tab_icon["log"] = $("<img src='images/console.png' class='toolbar_icon'>");
	}
	
	// whether icon should be shown on main toolbar
	this.ready = function() {
		return admin_check();
	}
	
	// try to open this panel (return false to cancel)
	this.show = function() {
		this.show_hidden_frames = false;
		this.visualcal_mode = false;
        targetrow.redraw();
		return true;
	}
	
	// try to close this panel (return false to cancel)
	this.hide = function() {
		this.visualcal_mode = false;
        targetrow.redraw();
		return true;
	}

	this.tab_current = get_cookie("system_tab_current") || "display"; // default tab name
	
	this.set_tab = function(tab) {
		this.tab_current = tab;
		set_cookie("system_tab_current", tab);
	}

	// display the panel in the supplied div
	this.draw = function(div) {
		if (toolbar_military.is_mil()) {
			this.tab_icon["users"] = $("<img src='images/user-mil.png' class='toolbar_icon'>");
		} else {
			this.tab_icon["users"] = $("<img src='images/user.png' class='toolbar_icon'>");
		}

		var frame = cur_string().frame();
		
		toolbar_system.draw_div_shutdown(div); // shutdown server

		// shot simulator toggle
		toolbar.new_button(div, S("ADMIN_SIMULATE"), function() {
			toolbar_system.simulate_mode = !toolbar_system.simulate_mode;
			toolbar_system.visualcal_mode = false;
			update();
		}, this.simulate_mode);
	
		// enable visual calibration
		toolbar.new_button(div, S("FRAME_VISUALCAL"), function() {
			toolbar_system.visualcal_mode = !toolbar_system.visualcal_mode;
			toolbar_system.simulate_mode = false;
			update();
		}, this.visualcal_mode);
		
		// fix date/time
		if (abs(cur_ts_local_error) > 1000 * 60 * 10) { // off by more than 10 minutes

			div.append("<br>");
			toolbar.new_text(div, S("ADMIN_SYNCTIME", round(cur_ts_local_error/60000)));
			
			toolbar.new_button(div, S("ADMIN_SYNCTIME_FIX"), function() {
				g_socket.emit("reset_time_now", Date.now());
			}).css("background", "#ff8");

		}
		toolbar.new_button(div, "INFO++", function() {
			toolbar_system.info_plus = !toolbar_system.info_plus;
			update();
		}, this.info_plus);
		if (g_config.Servertemperature && 'number'== typeof g_config.Servertemperature){
			toolbar.new_text(div, /*"server temp="*/ ""+g_config.Servertemperature+'C'); // just the temp, don't draw attention
		}

		
		// tab icon bar
		var row = $("<div>");
		row.css("background-color", "#888");
		row.css("margin-bottom", "12px");
		for (var tab_name in this.tab_icon) {
			if (tab_name == "gpios" && !g_config.is_JRPL) continue; // only draw "gpios" for JRPL
			if (tab_name == "rstat" && !g_config.is_rstat) continue; // turn off RSTAT for most
			var icon = this.tab_icon[tab_name];
			icon[0].arg_name = tab_name;
			icon.mousedown(function() {
				toolbar_system.set_tab(this.arg_name);
				update();
			});
			row.append(icon);
			icon.css("padding", "3px");
			if (toolbar_system.tab_current == tab_name) icon.css("border", "2px solid");
			else icon.css("border", "0");
		}
		div.append(row);
		
		if (this.tab_current == "display") {
		
			// display options
			toolbar.new_text(div, S("ADMIN_OPTIONS_DISPLAY"));
	
			// display group stats
			toolbar.new_button(div, S("ADMIN_OPTIONS_DISPLAY_STATS"), function() {
				config_crud({ option_group_stats: !g_config.option_group_stats });
			}, g_config.option_group_stats);
	
			// display shot circle
			toolbar.new_button(div, S("ADMIN_OPTIONS_DISPLAY_CIRCLE"), function() {
				config_crud({ option_shot_circle: !g_config.option_shot_circle });
			}, g_config.option_shot_circle);
	
			// display center of group cross
			toolbar.new_button(div, S("ADMIN_OPTIONS_DISPLAY_CENTER"), function() {
				config_crud({ option_group_center: !g_config.option_group_center });
			}, g_config.option_group_center);
	
			// display shot's x,y coords to users
			toolbar.new_button(div, S("ADMIN_OPTIONS_DISPLAY_SHOTXY"), function() {
				config_crud({ option_shot_xy: !g_config.option_shot_xy });
			}, g_config.option_shot_xy);
	
			// debug/devel purposes - extended stats
			/* if (g_config.is_JRPL)*/ toolbar.new_button(div, "XtraStats", function() {
				config_crud({ option_xtrastats: !g_config.option_xtrastats });
			}, g_config.option_xtrastats);
	
	
			// unit preferences
			
			div.append("<br>");
			toolbar.new_text(div, S("ADMIN_OPTIONS_COORDS"));
	
			toolbar.new_button(div, S("ADMIN_OPTIONS_COORDS_MOA"), function() {
				config_crud({ option_coords: "moa" });
			}, g_config.option_coords == "moa").addClass("button_small");
	
			toolbar.new_button(div, S("ADMIN_OPTIONS_COORDS_MIL"), function() {
				config_crud({ option_coords: "mil" });
			}, g_config.option_coords == "mil").addClass("button_small");
	
			toolbar.new_button(div, S("ADMIN_OPTIONS_COORDS_INCH"), function() {
				config_crud({ option_coords: "inch" });
			}, g_config.option_coords == "inch").addClass("button_small");
	
			toolbar.new_button(div, S("ADMIN_OPTIONS_COORDS_MM"), function() {
				config_crud({ option_coords: "mm" });
			}, g_config.option_coords == "mm").addClass("button_small");
	
			div.append("<br>");
			toolbar.new_text(div, S("ADMIN_OPTIONS_SPEED"));

			toolbar.new_button(div, S("ADMIN_OPTIONS_SPEED_FPS"), function() {
				config_crud({ option_speed_unit: "fps" });
			}, g_config.option_speed_unit == "fps").addClass("button_small");

			toolbar.new_button(div, S("ADMIN_OPTIONS_SPEED_MS"), function() {
				config_crud({ option_speed_unit: "m/s" });
			}, g_config.option_speed_unit == "m/s").addClass("button_small");
	
			div.append("<br>");
			
			// GRID stuff was here; moved to viewer_toolbar_info.js

			// select language
			div.append("<br>");
			toolbar.new_text(div, S("ADMIN_OPTIONS_LANGUAGE"));

			var languages = getLanguages();
			_.forEach( languages, function( lang ) {
				var item = toolbar.new_button(div, lang.label, function() {
					config_crud({ language: this.arg });
				}, g_config.language == lang.key ).addClass( "button_small" );
				item[ 0 ].arg = lang.key;
			} );
			
		} else if (this.tab_current == "frame") {

			toolbar.new_text(div, frame.data.name + ":");
			
			// enable/disable this target frame
			toolbar.new_button(div, (frame.data.enabled ? S("FRAME_HIDE") : S("FRAME_SHOW")), function() {
				g_socket.emit("frame_set_enabled", frame.id(), !frame.data.enabled);
			});
	
			toolbar.new_button(div, S("FRAME_RENAME"), function() {
				var name = prompt(S("FRAME_RENAME_PROMPT", frame.data.name), frame.data.name);
				if (name) {
					frame.crud({name: name});
				}
			});
	
			toolbar.new_button(div, S("FRAME_DELETE"), function() {
				if (numberof(g_frames) == 1) {
					alert(S("FRAME_DELETE_LAST"));
					return;
				}
				var res = confirm(S("FRAME_DELETE_CONFIRM", frame.data.name));
				if (res) {
					g_socket.emit("frame_delete", frame.id());
				}
			});
			
			// new frame
			toolbar.new_button(div, "Copy", function() { // todo
				var res = confirm(S("FRAME_NEW_CONFIRM", frame.data.name));
				if (res) {
					g_socket.emit("frame_create", frame.id()); // as copy of current frame
				}
			});

	
			if (frame.data.enabled) {
// DMC 3 May 2017... add support for alternate sensor geometries
				var button_vals =  ["STD"//, "STDNR", "STDMN"
                    //,"IMPACT"
                    ,"HBAR4","HBAR5"/*,"HBAR5A","HBAR5K"*/,"HBAR6"
                    //,"HVBAR5", "HVBAR6"/*,"CHAMBER"*/
                    ,"VBAR5","VBAR6"
                ];
				if (-1 == button_vals.indexOf(frame.data.geom.sens_geom)) {
					frame.crud({geom: {sens_geom:"STD"}});
				}

				if (g_config.nonstd_sensor_geom || frame.data.geom.sens_geom != "STD") {
			 		div.append("<br>");
					toolbar.new_text(div, "Sensor Geometry"); // TODO: S("...")
					for (var button_idx in button_vals) {
						(function (button_txt) {
							toolbar.new_button(div, button_txt, function() {
console.log("smeg2",button_txt);

								frame.crud({geom: {sens_geom:button_txt}});
							},frame.data.geom.sens_geom == button_txt);
						})(button_vals[button_idx]);
					}
				}
// DMC 9 Nov 2017... radio-button selection of standard sensor wiring
		 		div.append("<br>");
				toolbar.new_text(div, "Wiring Harness"); // TODO: S("...")
				var button_vals =  ["SOLO","G2-5mic","G2-8mic", "PnP","Custom"];
				if (-1 == button_vals.indexOf(frame.data.geom.wiring_harness)) {
					frame.crud({geom: {wiring_harness:"Custom"}});
				}

				for (var button_idx in button_vals) {
					(function (button_txt) {
						toolbar.new_button(div, button_txt, function() {
							frame.crud({geom: {wiring_harness:button_txt}});
							if (button_txt == "G2-5mic") frame.crud({geom: {chrono_spacing: 200}});
							else if (button_txt == "SOLO") frame.crud({geom: {chrono_spacing: 170}});
							else if (button_txt != "Custom") frame.crud({geom: {chrono_spacing: 210}});
							if (button_txt == "SOLO") frame.crud({geom: {mic_mapping: "58472613"}});
							if (button_txt == "G2-5mic") frame.crud({geom: {mic_mapping: "12345"}});
							if (button_txt == "G2-8mic") frame.crud({geom: {mic_mapping: "12345678"}});
							if (button_txt == "PnP") frame.crud({geom: {mic_mapping: "12365487"}});
						},frame.data.geom.wiring_harness == button_txt).addClass("button_small");
					})(button_vals[button_idx]);
				}

		 		div.append("<br>");
		 		
				var geom_is_std = "STD" == frame.data.geom.sens_geom;
				var geom_is_std = 0 == frame.data.geom.sens_geom.indexOf("STD");
				var geom_is_chamber = "CHAMBER" == frame.data.geom.sens_geom;
				var geom_is_hbar = 0 == frame.data.geom.sens_geom.indexOf("HBAR");
				var geom_is_hvbar = 0 == frame.data.geom.sens_geom.indexOf("HVBAR");
				var geom_is_vbar = 0 == frame.data.geom.sens_geom.indexOf("VBAR");
				if (geom_is_std || geom_is_chamber) {
				toolbar.new_text(div, S("FRAME_MICWH"));
				toolbar.new_button(div, round(frame.data.geom.mic_width) + ", " + round(frame.data.geom.mic_height), function() {
					var result = prompt(S("FRAME_MICWH_PROMPT", round(frame.data.geom.mic_width), round(frame.data.geom.mic_height)));
					if (result) {
						var cleaned = result.replace( /[^0-9 -.]/g, '').replace( /\s\s+/g, ' ' ).trim();
						var vals = cleaned.split(" ");
						if (vals.length == 2) {
							var mx = Number(vals[0]);
							var my = Number(vals[1]);
							if (!isNaN(mx) && !isNaN(my)) {
								frame.crud({geom: {mic_width: mx, mic_height: my}});
							}
						}
					}
				}).addClass("button_small");
				}
				
				if (geom_is_std || geom_is_hbar || geom_is_vbar) {
		 		toolbar.new_text(div, S("FRAME_SPACING"));
				toolbar.new_button(div, round(frame.data.geom.chrono_spacing), function() {
					var result = prompt(S("FRAME_SPACING_PROMPT"));
					if (result) {
						var spacing = Number(result.trim());
						if (!isNaN(spacing)) {
							frame.crud({geom: {chrono_spacing: spacing}});
						}
					}
				}).addClass("button_small");
				}

				if(g_config.nonstd_sensor_geom) {
					toolbar.new_text(div, "psi");
					var psi = frame.data.geom.psi || 0;frame.data.geom
					toolbar.new_button(div, psi, function() {
						var result = prompt("Enter rotation of sensors, positive = clockwise with respect to targetface");
						if (result) {
							var res = Number(result.trim());
							if (!isNaN(res)) {
								frame.crud({geom: {psi: trim_num(res,1)}});
							}
						}
					}).addClass("button_small");
				}
				
		 		div.append("<br>");

				if (geom_is_hbar || geom_is_vbar) {
					var vals =  ["X13","X26","X47","X58"];
                    // if (geom_is_vbar) vals.push("DY4");
                    if (geom_is_vbar) vals.push("DY1");
                    if (geom_is_vbar) vals.push("DY5");
                    vals.push("Y0");
					for (var idx in vals) {
						(function (txt) {
					 		toolbar.new_text(div, txt);
							toolbar.new_button(div, round(frame.data.geom[txt]), function() {
								var result = prompt(txt);
								if (result) {
									var cleaned = result.replace( /[^0-9 -.]/g, '').replace( /\s\s+/g, ' ' ).trim();
									var xx = Number(cleaned);
									if (!isNaN(xx) ) {
										var crudmsg = {geom:{}};
										crudmsg.geom[txt] = xx;
										frame.crud(crudmsg);
									}
								}
							}).addClass("button_small");
						})(vals[idx]);
					}
				}
				if (geom_is_hvbar) {
					var vals =  ["X136","X2","X4","X587","Y0"];
					for (var idx in vals) {
						(function (txt) {
					 		toolbar.new_text(div, txt);
							toolbar.new_button(div, round(frame.data.geom[txt]), function() {
								var result = prompt(txt);
								if (result) {
									var cleaned = result.replace( /[^0-9 -.]/g, '').replace( /\s\s+/g, ' ' ).trim();
									var xx = Number(cleaned);
									if (!isNaN(xx) ) {
										var crudmsg = {geom:{}};
										crudmsg.geom[txt] = xx;
										frame.crud(crudmsg);
									}
								}
							}).addClass("button_small");
						})(vals[idx]);
					}
				}

				if (geom_is_hvbar) {
					var vals =  ["DY6","DY7"];
					for (var idx in vals) {
						(function (txt) {
					 		toolbar.new_text(div, txt);
							toolbar.new_button(div, round(frame.data.geom[txt]), function() {
								var result = prompt(txt);
								if (result) {
									var cleaned = result.replace( /[^0-9 -.]/g, '').replace( /\s\s+/g, ' ' ).trim();
									var xx = Number(cleaned);
									if (!isNaN(xx) ) {
										var crudmsg = {geom:{}};
										crudmsg.geom[txt] = xx;
										frame.crud(crudmsg);
									}
								}
							}).addClass("button_small");
						})(vals[idx]);
					}
				}
		 		div.append("<br>");
		 		
				// display calibration values
		 		toolbar.new_text(div, S("FRAME_CALXY"));
				toolbar.new_button(div, round(frame.data.geom.x_cal) + ", " + round(frame.data.geom.y_cal), function() {
					var result = prompt(S("FRAME_CALXY_PROMPT", round(frame.data.geom.x_cal), round(frame.data.geom.y_cal)));
					if (result) {
						var cleaned = result.replace( /[^0-9 -.]/g, '').replace( /\s\s+/g, ' ' ).trim();
						var vals = cleaned.split(" ");
						if (vals.length == 2) {
							var mx = Number(vals[0]);
							var my = Number(vals[1]);
							if (!isNaN(mx) && !isNaN(my)) {
	
								var prev_x_cal = frame.data.geom.x_cal;
								var prev_y_cal = frame.data.geom.y_cal;
								var new_x_cal = mx;
								var new_y_cal = my;
	
								var crud = new_crud(); // build message to send to server
								crud.frames[frame.id()] = {};
								crud.strings[cur_string().id()] = {};
	
								// apply new offsets
								crud.frames[frame.id()].geom = { x_cal: new_x_cal, y_cal: new_y_cal };
	
								crud.strings[cur_string().id()].shots = [];
	
								// update all existing shots in string
								var shots = cur_string().shots();
								for (var i in shots) {
									var shot = shots[i];
	
									// new position
									var new_x = shot.x - prev_x_cal + new_x_cal;
									var new_y = shot.y - prev_y_cal + new_y_cal;
									var new_rawX = shot.rawX - prev_x_cal + new_x_cal;
									var new_rawY = shot.rawY - prev_y_cal + new_y_cal;
	
									// update on server
									crud.strings[cur_string().id()].shots[i] = {
										x: new_x,
										y: new_y,
										rawX: new_rawX,
										rawY: new_rawY,
										geom: {
											x_cal: new_x_cal,
											y_cal: new_y_cal
										}
									};
	
								}
	
								// send this request to server
								// do not also update client-side - better to be safe
								g_socket.emit("crud", crud);
							}
						}
					}
				}).addClass("button_small");
	
		 		toolbar.new_text(div, S("FRAME_TEMP"));
				toolbar.new_button(div, round(frame.data.geom.temp_cal), function() {
					var result = prompt(S("FRAME_TEMP_PROMPT"));
					if (result) {
						var tcal = Number(result.trim());
						if (!isNaN(tcal)) {
							frame.crud({geom: {temp_cal: round(tcal)}});
						}
					}
				}).addClass("button_small");
				
				if (g_config.enable_angle_cal) {
	
			 		div.append("<br>");
			 		
					toolbar.new_button(div, "algo-Bisley", function() {
										frame.crud({geom: {bisley_behaviour: !frame.data.geom.bisley_behaviour}});
					}, frame.data.geom.bisley_behaviour).addClass("button_small");

			 		toolbar.new_text(div, S("FRAME_THETA"));
			 		var theta = frame.data.geom.theta_cal || 0;
					toolbar.new_button(div, theta, function() {
						var result = prompt(S("FRAME_THETA_PROMPT"));
						if (result) {
							var res = Number(result.trim());
							if (!isNaN(res)) {
								frame.crud({geom: {theta_cal: round(res)}});
							}
						}
					}).addClass("button_small");
			 		toolbar.new_text(div, '+/-');
			 		var theta = frame.data.geom.theta_window || 0;
					toolbar.new_button(div, theta, function() {
						var result = prompt();
						if (result) {
							var res = Number(result.trim());
							if (!isNaN(res) && (res=round(res)) >= 0) {
								frame.crud({geom: {theta_window: res}});
							}
						}
					}).addClass("button_small");

			 		toolbar.new_text(div, ',&nbsp;&nbsp;');
			 		toolbar.new_text(div, S("FRAME_PHI"));
			 		var phi = frame.data.geom.phi_cal || 0;
					toolbar.new_button(div, phi, function() {
						var result = prompt(S("FRAME_PHI_PROMPT"));
						if (result) {
							var res = Number(result.trim());
							if (!isNaN(res)) {
								frame.crud({geom: {phi_cal: round(res)}});
							}
						}
					}).addClass("button_small");
			 		toolbar.new_text(div, '+/-');
			 		var phi = frame.data.geom.phi_window || 0;
					toolbar.new_button(div, phi, function() {
						var result = prompt();
						if (result) {
							var res = Number(result.trim());
							if (!isNaN(res) && (res=round(res)) >= 0) {
								frame.crud({geom: {phi_window: res}});
							}
						}
					}).addClass("button_small");
/*					
			 		if ("HBAR5K" == frame.data.geom.sens_geom) {
				 		toolbar.new_text(div, ',&nbsp;&nbsp;K1');
				 		var k1 = frame.data.geom.k1 || 0;
						toolbar.new_button(div, k1, function() {
							var result = prompt();
							if (result) {
								var res = Number(result.trim());
								if (!isNaN(res)) {
									frame.crud({geom: {k1: res}});
								}
							}
						}).addClass("button_small");
					}
			 		if ("HBAR5A" == frame.data.geom.sens_geom) {
				 		toolbar.new_text(div, ',&nbsp;&nbsp;A1');
				 		var a1 = frame.data.geom.a1 || 0;
						toolbar.new_button(div, a1, function() {
							var result = prompt();
							if (result) {
								var res = Number(result.trim());
								if (!isNaN(res)) {
									frame.crud({geom: {a1: res}});
								}
							}
						}).addClass("button_small");
					}
*/			 		

				} // if enable_angle_cal
				
				div.append("<br>");
				toolbar.new_text(div, S("ADMIN_MICMAPPING"));
				
				var str = "12345";
				if (g_config.mic_mapping) str = g_config.mic_mapping;
				if (frame.data.geom.mic_mapping) str = frame.data.geom.mic_mapping;
				// custom ?
				var display_fn = toolbar.new_text;
				if (frame.data.geom.wiring_harness == 'Custom') display_fn = toolbar.new_button;
				// sensor mapping editable only if "Custom"
				display_fn(div, str, function() {
				// toolbar.new_button(div, str, function() {
					var result = prompt(S("ADMIN_MICMAPPING_PROMPT"));
					if (result == null) return; // cancel
					result = result.replace( /[^0-8]/g, ''); // accept only valid digits
					// allow empty string, which means revert to default
					if (result == "" || (result.length >= 4 && result.length <= 8)) {
						frame.crud({geom: {mic_mapping: result}});
					} else {
						alert(S("ADMIN_MICMAPPING_INVALID"));
					}
				}).addClass("button_small");
	
				// frame association to targets
				div.append("<hr>");
	
				if (frame.device()) { // connected to a device
					var device_data = frame.device().data;
	
					var str = S("FRAME_CONNECTED", frame.device().device_type_and_short_name());
					if (device_data.batteryvalue >= 0) str += S("FRAME_BATTERY", device_data.batteryvalue);
					toolbar.new_text(div, str);
					
					// disconnect button (todo: only needed for servers)
					toolbar.new_button(div, S("FRAME_DISCONNECT"), function() {
						g_socket.emit("frame_unassociate", frame.id());
					});
					
					// only useful with fw=103, which we are unlikely to use
					if (false && device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
						var device_id = frame.device().id();
						toolbar.new_button(div, "RESET", function() {
							var crud = { devices: {} }; // build crud message
							crud.devices[device_id] = {SOLO_reset: !device_data.SOLO_reset};
							g_socket.emit("crud", crud);
						}, device_data.SOLO_reset).addClass("button_small");
					}

					if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
						var device_id = frame.device().id();
						toolbar.new_button(div, "TSOLO-sw-update", function() {
							var crud = { devices: {} }; // build crud message
							crud.devices[device_id] = {SOLO_C2_sw_update: !device_data.SOLO_C2_sw_update};
							g_socket.emit("crud", crud);
							g_socket.emit("SOLO_C2_sw_update", {device_id: device_id});
						}, device_data.SOLO_C2_sw_update).addClass("button_small");
					}
					if (0) { // only for SW dev and testing
						if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
							var device_id = frame.device().id();
							toolbar.new_button(div, "C2-sw2", function() {
								var crud = { devices: {} }; // build crud message
								crud.devices[device_id] = {SOLO_C2_sw_update2: !device_data.SOLO_C2_sw_update2};
								g_socket.emit("crud", crud);
								g_socket.emit("SOLO_C2_sw_update2", {device_id: device_id});
							}, device_data.SOLO_C2_sw_update2).addClass("button_small");
						}

						if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
							var device_id = frame.device().id();
							toolbar.new_button(div, "R2:heavy", function() {
								var crud = { devices: {} }; // build crud message
								crud.devices[device_id] = {SOLO_C2_reset2_heavy: !device_data.SOLO_C2_reset2_heavy};
								g_socket.emit("crud", crud);
								if (crud.devices[device_id].SOLO_C2_reset2_heavy) {
									g_socket.emit("SOLO_C2_reset2_heavy", {device_id: device_id});
								}
							}, device_data.SOLO_C2_reset2_heavy).addClass("button_small");
						}

						if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
							var device_id = frame.device().id();
							toolbar.new_button(div, "R2:medium", function() {
								var crud = { devices: {} }; // build crud message
								crud.devices[device_id] = {SOLO_C2_reset2_medium: !device_data.SOLO_C2_reset2_medium};
								g_socket.emit("crud", crud);
								if (crud.devices[device_id].SOLO_C2_reset2_medium) {
									g_socket.emit("SOLO_C2_reset2_medium", {device_id: device_id});
								}
							}, device_data.SOLO_C2_reset2_medium).addClass("button_small");
						}

						if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
							var device_id = frame.device().id();
							toolbar.new_button(div, "RESET:medium", function() {
								var crud = { devices: {} }; // build crud message
								crud.devices[device_id] = {SOLO_C2_reset_medium: !device_data.SOLO_C2_reset_medium};
								g_socket.emit("crud", crud);
								if (crud.devices[device_id].SOLO_C2_reset_medium) {
									g_socket.emit("SOLO_C2_reset_medium", {device_id: device_id});
								}
							}, device_data.SOLO_C2_reset_medium).addClass("button_small");
						}

						if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
							var device_id = frame.device().id();
							toolbar.new_button(div, "RESET:light", function() {
								var crud = { devices: {} }; // build crud message

								crud.devices[device_id] = {SOLO_C2_reset_light: !device_data.SOLO_C2_reset_light};
								g_socket.emit("crud", crud);
								if (crud.devices[device_id].SOLO_C2_reset_light) {
									g_socket.emit("SOLO_C2_reset_light", {device_id: device_id});
								}
							}, device_data.SOLO_C2_reset_light).addClass("button_small");
						}
					} else {
						if (device_data.type == 'SOLO' /* && g_config.is_fwtweaks !== false */) {
							var device_id = frame.device().id();
							toolbar.new_button(div, "SOLO-soft-reset", function() {
								var crud = { devices: {} }; // build crud message
								crud.devices[device_id] = {SOLO_C2_reset2_medium: !device_data.SOLO_C2_reset2_medium};
								g_socket.emit("crud", crud);
								if (crud.devices[device_id].SOLO_C2_reset2_medium) {
									g_socket.emit("SOLO_C2_reset2_medium", {device_id: device_id});
								}
							}, device_data.SOLO_C2_reset2_medium).addClass("button_small");
						}

					}
					// SOLO hwversion/fwversion
					
					var hwversion = device_data && device_data.hwversion;
					var fwversion = device_data && device_data.fwversion;
					var c2fwversion = device_data && device_data.c2fwversion;
					if (fwversion || hwversion || c2fwversion) {
						var txt="";
						if (fwversion) {
							txt += (txt!="") ? "," : "";
							txt += "F="+fwversion;
						}
						if (hwversion) {
							txt += (txt!="") ? "," : "";
							txt += "H="+hwversion;
						}
						if (c2fwversion) {
							txt += (txt!="") ? "," : "";
							txt += c2fwversion;
						}
						toolbar.new_text(div, "SOLO " + txt);
					}
					var ping_health = device_data && device_data.ping_health;
					if (ping_health && g_config.is_health) {
						var health_div = $('<span title="sme">')
							.text("Quality="+ping_health.show)
							.attr("title",JSON.stringify(ping_health,null,2));
						div.append(health_div);
					}
					
					// check most recent shot for invalid temperature reading
					if (cur_string().data.shots.length > 0) {
						var recent_shot = cur_string().data.shots[cur_string().data.shots.length - 1];
						if (recent_shot.temp_invalid) {
							toolbar.new_button(div, S("FRAME_TEMP_WARNING"), function() {
								alert(S("FRAME_TEMP_WARNING_MESSAGE"));
							}).css("background", "#f88");
						}
					}
	
				// only real servers can directly connect to anything
				} else if (g_config.device_type == "S10" || g_config.device_type == "S50") {
	
					// assign to device (new method using buttons)
					toolbar.new_text(div, S("FRAME_CONNECT"));
	
					for (i in g_devices) {
						var device = g_devices[i];
						if (!device.data.allow_association) continue;
						if (device.data.is_present == false) continue; // only if is_present or is not set (G1)
	
						var item = toolbar.new_button(div, device.device_type_and_short_name(), function() {
							g_socket.emit("frame_associate", frame.id(), this.arg);
						});
						item[0].arg = device.id();
						if (!device.frame()) item.addClass("button_blue");
					}
	
					toolbar.new_button(div, S("FRAME_CONNECT_OTHER"), function() {
						var res = prompt(S("FRAME_CONNECT_OTHER_PROMPT"));
						if (res) {
							var num = Number(res);
							if (!isNaN(num) && num >= 50 && num <= 150) {
								var ip = "192.168.0." + num;
								g_socket.emit("frame_associate", frame.id(), ip);
							} else {
								alert("invalid value " + num);
							}
						}
					});
	
					// provide button for directly connecting to G2 if enabled by server
					if (g_config.other_G2_button) {
						toolbar.new_button(div, S("FRAME_CONNECT_OTHER_G2"), function() {
							var res = prompt(S("FRAME_CONNECT_OTHER_PROMPT"));
							if (res) {
								var num = Number(res);
								if (!isNaN(num) && num >= 50 && num <= 150) {
									var ip = "192.168.0." + num;
									g_socket.emit("frame_associate_G2", frame.id(), ip);
								} else {
									alert("invalid value " + num);
								}
							}
						});
					}
	
				// G2 can only connect to self
				} else if (g_config.device_type == "G2") {
					
					toolbar.new_text(div, S("FRAME_CONNECT_SELF"));
	
					toolbar.new_button(div, frame.data.name, function() {
						for (var i in g_devices) {
							if (g_devices[i].data.allow_association) {
								g_socket.emit("frame_associate", frame.id(), i);
							}
						}
					});
					
				}
				
			}

		} else if (this.tab_current == "faces") {
			
			// new import faces method (accepts js and tgz)
			var form = $("<form action='import_targetfaces' method='post' enctype='multipart/form-data' data-ajax='false' class='title_element'>");
			form.text(S("ADMIN_FACE_IMPORT"));
			var input = $("<input type='file' name='selected_file'>"); // multiple?
			glob_input.change(function(e) {
				if (confirm(S("ADMIN_FACE_IMPORT_CONFIRM"))) {
					this.form.submit();
				}
			});
			form.append(glob_input);
			div.append(form);
			
			// display list of faces to make active
			div.append("<br>");
			toolbar.new_text(div, S("ADMIN_FACE_ACTIVE") + " (hold to delete)");
			
			var inner = $("<div>");
			div.append(inner);

			inner.css("border", "1px solid #aaa");
			inner.css("overflow-y", "scroll");
			inner.css("max-height", maintarget.div.height() * 0.5);

			var face_list = [];
			for (var i in g_faces) {
				face_list.push(i);
			}

			// sort by distance
			face_list.sort(function(a, b) {
				if (g_faces[a].data.distance < g_faces[b].data.distance) return -1;
				if (g_faces[a].data.distance > g_faces[b].data.distance) return 1;
				return 0;
			});

			for (var i in face_list) {
				var face = g_faces[face_list[i]];
				var is_fclass = face.data.scoringzones_f;
				var item = toolbar.new_button(inner, face.data.name, function() {
					var face = g_faces[this.arg];
					face.crud({enabled: !face.data.enabled});
				}, face.data.enabled);
				item[0].arg = face_list[i];
				if (is_fclass) item.css("text-decoration", "underline");
				
				new Hammer(item[0]).on("press", function(e) {
					var face = g_faces[e.target.arg];
					if (confirm("Are you sure you want to delete targetface: " + face.data.name)) {
						g_socket.emit("face_delete", face.id());
					}
				});
				
				// buttons are fixed size and compact
				item.css("font-size", "14px");
				item.css("line-height", "12px");
				item.css("width", "75px");
				item.css("height", "50px");
				item.css("padding", "2px");
				item.css("margin", "2px");
				item.css("vertical-align", "middle");
				item.css("word-wrap", "break-word");

			}

		} else if (this.tab_current == "users") {

			// there are X range officers (click for list)
			var ro_list = [];
			for (var i in g_users) {
				if (g_users[i].data.name == "admin") continue;
				if (g_users[i].data.is_range_officer) ro_list.push(i);
			}
			
			toolbar.new_text(div, "There are " + ro_list.length + " "+ toolbar_military.stringname_RangeOfficers());
			if (ro_list.length > 0) {
				toolbar.new_button(div, "List", function() {
					var str = toolbar_military.stringname_RangeOfficers() + ":\n\n";
					for (var i in ro_list) {
						str += user_display_name(ro_list[i]) + "\n";
					}
					alert(str);
				});
			}
			
			div.append("<hr>");
			toolbar_system.drawUsersEditor(div);

		} else if (this.tab_current == "storage") {
			
			toolbar.new_text(div, S("REVIEW_SHOTSLOG"));

			// download shotslog
			toolbar.new_button(div, S("REVIEW_SHOTSLOG_4DAY"), function() {
				g_socket.emit("download_shotslog", 4); // number of days prior to most recent string
			});

			toolbar.new_button(div, S("REVIEW_SHOTSLOG_FULL"), function() {
				g_socket.emit("download_shotslog");
			});
			
			div.append("<br>");
		
			// disk space management and backup/restore
	
			var free_shots = S("ADMIN_STORAGE_OVER100K");
			if (g_config.diskspace < 100000) free_shots = g_config.diskspace;
			toolbar.new_text(div, S("ADMIN_STORAGE_REMAINING", free_shots));
	
			// delete all shot history
			toolbar.new_button(div, S("ADMIN_STORAGE_CLEAR"), function() {
				var res = prompt(S("ADMIN_STORAGE_CLEAR_PROMPT"));
				if (res) {
					var days = Number(res.trim());
					if (!isNaN(days) && days >= 0) {
						var res = confirm(S("ADMIN_STORAGE_CLEAR_CONFIRM", days));
						if (res) {
							g_socket.emit("clear_old_strings", days);
							update();
						}
					}
				}
			});
			
			// option to save invalid shots to string
			div.append("<br> invalid sensor readings");

			// option to purge invalid shots from open stings
			toolbar.new_button(div, "auto purge", function() {
				config_crud({ autopurge_invalid_shots: !g_config.autopurge_invalid_shots });
			}, g_config.autopurge_invalid_shots).addClass("button_small");

			// backup all data to tgz file
			div.append("<br>");
			toolbar_system.draw_div_backup(div);
			toolbar_system.draw_div_restore(div);

if (0) {			
			div.append($('<hr>'));

			
			var csvform = $("<form action='import-csv' method='post' enctype='multipart/form-data' data-ajax='false' class='title_element'>");
			csvform.text('import csv file: ');

		var loc_input = $("<input type='file' name='selected_file'>"); // multiple?

			loc_input.change(function(e) {
				if (confirm('import CSV?')) {
					this.form.submit();
				}
			});
			
			csvform.append(loc_input);
			div.append(csvform);

			div.append($('<hr>EXPORT CSV:</hr>'));

			toolbar.new_button(div, "users-csv")//.addClass("button_small")
				.wrapInner($('<a class=linkbutton href="/download-csv/users"/>'))
				.addClass("button_small")


			div.append($('<a href="/download-csv/users" sdownload > ABC </a>'));


//					td.wrapInner($('<a style="text-decoration: none" href="http://' + device.data.ip + '"/>'));
}

			div.append("</td></tr></table>");

		} else if (this.tab_current == "advanced") {
			
			// links to advanced things
//			var item = toolbar.new_text(div, S("ADMIN_ADVANCED_LINKS"));
			
			// // window.location.hostname = "localhost"; window.location.host = "localhost:4004"
			// var link = $("<a href='http://" + window.location.hostname + ":280' target='_blank'>");
			// toolbar.new_button(link, S("ADMIN_ADVANCED_LINKS") + S("ADMIN_ADVANCED_LINKS_280"));
			// div.append(link);

			// var link = $("<a href='http://" + window.location.host + "/stats.html' target='_blank'>");
			// toolbar.new_button(link, "PingStats");
			// div.append(link);

//			div.append("<br>");
			
			toolbar.new_button(div, "Refresh All Clients", function() {
				if (confirm("Force all clients to reload the page?"))
					g_socket.emit("client_refresh");
			});
			
			toolbar.new_button(div, "Software Restart", function() {
				if (confirm("Stop the server process? It should automatically restart in 5 seconds."))
					g_socket.emit("stop_process");
			});

			toolbar.new_button(div, "Reboot", function() {
				if (confirm("Reboot the server entirely? It should automatically restart in 1 minute."))
					g_socket.emit("reboot_server");
			});

			if (g_config.version_string) {
				div.append("<br>");
				toolbar.new_text(div, S("ADMIN_ADVANCED_VERSION", g_config.version_string));
			}
	
			div.append("<br> TARGETFRAME options");
			var txt = "frozen side/plunge angles"; // TODO: must update S("ADMIN_ENABLEANGLECAL")
			toolbar.new_button/*new_checkbox*/(div, txt, function() {
				config_crud({ enable_angle_cal: !g_config.enable_angle_cal });
			}, g_config.enable_angle_cal).addClass("button_small");

			var txt = "nonstandard sensor geom"; // TODO: must create S("ADMIN_xxxx")
			toolbar.new_button/*new_checkbox*/(div, txt, function() {
				config_crud({ nonstd_sensor_geom: !g_config.nonstd_sensor_geom });
			}, g_config.nonstd_sensor_geom).addClass("button_small");

			// div.append("<br>");
			// var txt = "Global sensor mapping"; // TODO: must update S("ADMIN_MICMAPPING")
			// toolbar.new_text(div, txt);
			
			// var str = g_config.mic_mapping;
			// if (!str) str = "12345";
			// toolbar.new_button(div, str, function(event) {
			// 	var result = prompt(S("ADMIN_MICMAPPING_PROMPT"));
			// 	if (result == null) return; // cancel
			// 	if (!result) result = "12345"; // clear to set default
			// 	result = result.replace( /[^0-8]/g, ''); // accept only valid digits
			// 	if (result.length >= 4 && result.length <= 8) {
			// 		config_crud({ mic_mapping: result });
			// 	} else {
			// 		alert(S("ADMIN_MICMAPPING_INVALID"));
			// 	}
			// });

			// var txt = "frame override"; // TODO: must update S("ADMIN_MICMAPPING_PERFRAME")
			// toolbar.new_button(div, txt, function() {
			// 	config_crud({ micmapping_perframe: !g_config.micmapping_perframe });
			// }, g_config.micmapping_perframe).addClass("button_small");
			
			div.append("<hr>");

			if (g_config.is_SMT) {
		 		toolbar.new_text(div, 'timers: ');

		 		toolbar.new_text(div, 'SpinningWheel(500)');
		 		var timeoutSpinner = g_config.option_timeout_spinner || 500;
				toolbar.new_button(div, timeoutSpinner, function() {
					var result = prompt();
					if (result) {
						var res = Number(result.trim());
						if (!isNaN(res) /*&& (res=round(res)) >= 0*/) {
							config_crud({ option_timeout_spinner: res });
						}
					}
				}).addClass("button_small");

		 		toolbar.new_text(div, 'Red-X(10000)');
		 		var timeoutRedX = g_config.option_timeout_red_x || 10000;
				toolbar.new_button(div, timeoutRedX, function() {
					var result = prompt();
					if (result) {
						var res = Number(result.trim());
						if (!isNaN(res) /*&& (res=round(res)) >= 0*/) {
							config_crud({ option_timeout_red_x: res });
						}
					}
				}).addClass("button_small");

				toolbar.new_checkbox(div, S("Ping"), function() {
					config_crud({ client_ping: !g_config.client_ping });
				}, g_config.client_ping);

				toolbar.new_button(div, "debuginfo", function() { // turn on extended (mostly JRPL) functions
					config_crud({ option_debuginfo: !g_config.option_debuginfo });
				}, g_config.option_debuginfo).addClass("button_small");

				div.append("<br>");
			}

	 		toolbar.new_text(div, 'recv_shot discard: ');
	 		
			toolbar.new_button(div, "tap", function() {
				config_crud({ recv_shot_reject_tap: !g_config.recv_shot_reject_tap }); // no-soln, single-sensor readings
			}, g_config.recv_shot_reject_tap).addClass("button_small");

			toolbar.new_button(div, "multi", function() {
				config_crud({ recv_shot_reject_multi: !g_config.recv_shot_reject_multi }); // no-soln, multi-sensor
			}, g_config.recv_shot_reject_multi).addClass("button_small");

			toolbar.new_button(div, "off-near", function() {
				config_crud({ recv_shot_reject_off_near: !g_config.recv_shot_reject_off_near }); // soln off-target but near
			}, g_config.recv_shot_reject_off_near).addClass("button_small");

			toolbar.new_button(div, "off-far", function() {
				config_crud({ recv_shot_reject_off_far: !g_config.recv_shot_reject_off_far }); // soln off-target - all!
			}, g_config.recv_shot_reject_off_far).addClass("button_small");


			div.append("<hr>");
	 		toolbar.new_text(div, 'recv_shot mark invalid: ');

			toolbar.new_button(div, "ghost-V", function() {
				config_crud({ recv_shot_invalidate_ghostV: !g_config.recv_shot_invalidate_ghostV }); // ghost shots (by speed)
			}, g_config.recv_shot_invalidate_ghostV).addClass("button_small");

			// mostly for H-BAR; invalidate shot if any chronos are negative
			toolbar.new_button(div, "vNeg", function() {
				config_crud({ star_shot_vneg: !g_config.star_shot_vneg });
			}, g_config.star_shot_vneg).addClass("button_small");

			div.append("<br>");
	 		toolbar.new_text(div, 'recv_shot invalid angles: ');
	 		// DEPRECATED: g_config.recv_shot_invalidate_angles

	 		toolbar.new_text(div, S("FRAME_THETA"));
	 		var theta = g_config.recv_shot_angle_theta_mean || 0;

			toolbar.new_button(div, theta, function() {
				var result = prompt(S("FRAME_THETA_PROMPT"));
				if (result) {
					var res = Number(result.trim());
					if (!isNaN(res)) {
						config_crud({recv_shot_angle_theta_mean: round(res)});
					}
				}
			}).addClass("button_small");
	 		toolbar.new_text(div, '+/-');
	 		var theta = g_config.recv_shot_angle_theta_window || 0;
			toolbar.new_button(div, theta, function() {
				var result = prompt();
				if (result) {
					var res = Number(result.trim());
					if (!isNaN(res) && (res=round(res)) >= 0) {
						config_crud({recv_shot_angle_theta_window: res});
					}
				}

			}).addClass("button_small");
	 		toolbar.new_text(div, S("FRAME_PHI"));
	 		var phi = g_config.recv_shot_angle_phi_mean || 0;
			toolbar.new_button(div, phi, function() {
				var result = prompt(S("FRAME_PHI_PROMPT"));
				if (result) {
					var res = Number(result.trim());
					if (!isNaN(res)) {
						config_crud({recv_shot_angle_phi_mean: round(res)});
					}
				}
			}).addClass("button_small");
	 		toolbar.new_text(div, '+/-');
	 		var phi = g_config.recv_shot_angle_phi_window || 0;
			toolbar.new_button(div, phi, function() {
				var result = prompt();
				if (result) {
					var res = Number(result.trim());
					if (!isNaN(res) && (res=round(res)) >= 0) {
						config_crud({recv_shot_angle_phi_window: res});
					}
				}
			}).addClass("button_small");

			div.append("<hr>");
	 		toolbar.new_text(div, '* shot rules: ');

			toolbar.new_button(div, "MissingMic", function() {
				config_crud({ star_shot_missing_mic: !g_config.star_shot_missing_mic });
			}, g_config.star_shot_missing_mic).addClass("button_small");

			toolbar.new_button(div, "xyErr", function() {
				config_crud({ star_shot_xyerr: !g_config.star_shot_xyerr });
			}, g_config.star_shot_xyerr).addClass("button_small");

			if (g_config.star_shot_xyerr) toolbar.new_button(div
			, 'undefined' == typeof g_config.star_shot_xyerr_amt ? 25 : g_config.star_shot_xyerr_amt
			, function() {
				var result = prompt();
				if (result) {
					var res = Number(result.trim());
					if (!isNaN(res) /*&& (res=round(res)) >= 0*/) {
						config_crud({ star_shot_xyerr_amt: res });
					}
				}
			}).addClass("button_small");

			toolbar.new_button(div, "vErr", function() {
				config_crud({ star_shot_verr: !g_config.star_shot_verr });
			}, g_config.star_shot_verr).addClass("button_small");

			if (g_config.star_shot_verr) toolbar.new_button(div
			, 'undefined' == typeof g_config.star_shot_verr_amt ? 1 : g_config.star_shot_verr_amt
			, function() {
				var result = prompt();
				if (result) {
					var res = Number(result.trim());
					if (!isNaN(res) /*&& (res=round(res)) >= 0*/) {
						config_crud({ star_shot_verr_amt: res });
					}
				}
			}).addClass("button_small");

/*
// TO BE IMPLEMENTED:

			toolbar.new_button(div, "thErr", function() {
				config_crud({ star_shot_th_err: !g_config.star_shot_th_err }); // theta
			}, g_config.star_shot_therr).addClass("button_small");

			toolbar.new_button(div, "phErr", function() {
				config_crud({ star_shot_ph_err: !g_config.star_shot_ph_err }); // phi
			}, g_config.star_shot_pherr).addClass("button_small");
*/




/*


			 		var phi = frame.data.geom.phi_window || 0;
					toolbar.new_button(div, phi, function() {
						var result = prompt();
						if (result) {
							var res = Number(result.trim());
							if (!isNaN(res) && (res=round(res)) >= 0) {
								frame.crud({geom: {phi_window: res}});
							}
						}
					}).addClass("button_small");




			toolbar.new_button(div, S("ADMIN_OPTIONS_SPEED_FPS"), function() {
				config_crud({ option_speed_unit: "fps" });
			}, g_config.option_speed_unit == "fps").addClass("button_small");
*/

			div.append("<hr>");


			
		} else if (this.tab_current == "log") {
			
			// display running log of all sensors that are firing
			
			var now = (new Date()).getTime() - cur_ts_local_error;
			
			var trigger_log = [];

			// all active strings
			for (var i in g_strings) {
				var str = g_strings[i];
				if (str.frame().data.enabled && str.frame().string().id() == str.id()) {
					
					// search shots in reverse order
					var shots = str.shots();
					for (var j = shots.length - 1; j >= 0; j--) {
						var shot = shots[j]; // might be zero
						if (!shot || /* shot.ts < now - 20000 || */ shot.ts > now + 1000) break;
						trigger_log.push({str_id: i, shot: shot});
					}
				}
			}
			
			toolbar.new_text(div, "Sensor Triggering Log: (" + trigger_log.length + " shots)");
			div.append("<br>");
			
			if (trigger_log.length > 0) {
			
				trigger_log.sort(function(a,b) {
				    if (a.shot.ts > b.shot.ts) return -1;
				    if (a.shot.ts < b.shot.ts) return 1;
				    return 0;
				});

				var log = $("<div>");
				log.css("overflow-y", "scroll");
				log.css("margin-left", "30px");
				log.css("max-height", maintarget.div.height() - 300);

				// allow touch scrolling only when scrolling is possible
				// to prevent scrolling the entire page itself
				log.bind("touchstart", function(e) {
					toolbar.log_touch_start = e.originalEvent.touches[0].pageY;
					e.originalEvent.allow_touchstart = true;
				});

				log.bind("touchmove", function(e) {
					var delta = e.originalEvent.touches[0].pageY - toolbar.log_touch_start;
					if (delta > 0 && log[0].scrollTop == 0) return;
					if (delta < 0 && log[0].scrollTop == log[0].scrollHeight - log[0].clientHeight) return;
					e.originalEvent.allow_touchmove = true;
				});
				
				var count = 0;
				var ts_last = 0;
			
				for (var i in trigger_log) {
					var str = g_strings[trigger_log[i].str_id];
					var shot = trigger_log[i].shot;

					//log.append("" + shot.ts - ts_last);
					
					if (shot.ts - ts_last < -500) log.append("<hr>");
					ts_last = shot.ts;
					
					var line = $("<div>");
					
					// info button
					var item = toolbar.new_button(line, "info", function() {
						var date = new Date(this.arg.ts);
						str = date.toLocaleTimeString() + "<br>";
						str += JSON.stringify(this.arg, null, "    ");
						alert(str);
					}, shot.ts > now - 5000);
					item.css("font-size", "14px");
					item.css("font-weight", "bold");
					item.css("min-width", "0");
					item[0].arg = shot;

					line.append(str.frame().data.name + (cur_string().shot_is_whack(shot) ? "*" : " "));
					line.append(shot.valid ? "==>" : "   ");
					
					// show mics that fired
					var mic_str = "";
					if (shot.mics) {
						for (var j = 0; j < shot.mics.length; j++) {
							var t = shot.mics[j];
							if (t == t && t >= 0) {
								mic_str += (j+1);
							}
						}
					}
					
					line.append("<span class='boldnumber'>" + mic_str + "</span>");
	
					// has solution
					if (shot.res && shot.v) {
						
						var str = "";
						
						var v_count = 0;
						if (shot.res.v13) v_count++;
						if (shot.res.v26) v_count++;
						if (shot.res.v47) v_count++;
						if (shot.res.v58) v_count++;

						//str = " (x,y: <b>" + shot.x + ", " + shot.y + "</b>, v: <b>" + shot.v + "</b> [" + v_count + "])";
						
						str = " (v: ";
						str += "<b>" + ((shot.res.raw_v13)?(shot.res.raw_v13.toFixed(0)):(" --- ")) + "</b>, ";
						str += "<b>" + ((shot.res.raw_v26)?(shot.res.raw_v26.toFixed(0)):(" --- ")) + "</b>, ";
						str += "<b>" + ((shot.res.raw_v47)?(shot.res.raw_v47.toFixed(0)):(" --- ")) + "</b>, ";
						str += "<b>" + ((shot.res.raw_v58)?(shot.res.raw_v58.toFixed(0)):(" --- ")) + "</b>)";
						
						line.append(str);
					}
	
					log.append(line);
					
					if (++count > 50) break;
				}
				
				div.append(log);
				
			}

		} else if (this.tab_current == "rstat") {

			toolbar.new_button(div, "RSTAT sync all", function() {
				rstat_log = [];
				rstat_string_action = {};
				g_socket.emit("rstat_sync");
			});
			
			if (rstat_log.length > 0) {
				var msglog = $("<div>");
				msglog.css("font-size", "12px");
				msglog.css("background-color", "#eee");
				msglog.css("color", "black");
				msglog.css("max-height", maintarget.div.height() - 300);
				msglog.css("overflow-y", "scroll");
				
				var line_by_string = {};
				
				for (i in rstat_log) {
					var line = rstat_log[i];
					
					// saving score
					if (line.string) {
						var id = line.string._id;
						
						// file by string
						if (!line_by_string[id]) line_by_string[id] = [];
						line_by_string[id].push(line);
						
					// just print messages
					} else {
						msglog.append(line.text + "<br>");
					}
					
				}
						
				function view_button(div, string_id) {
					toolbar.new_button(div, "View", function() {
						g_socket.emit("get_string", {string_id: string_id});
					}, g_string_view && g_string_view.id() == string_id).addClass("button_small");
				}
				
				function action_button(div, string_id, text, action) {
					var item = toolbar.new_button(div, text, function() {
						if (rstat_string_action[string_id] != action) {
							g_socket.emit("rstat_string_action", string_id, action);
							rstat_string_action[string_id] = action;
						} else {
							g_socket.emit("rstat_string_action", string_id, "none");
							rstat_string_action[string_id] = "none";
						}
						toolbar.update();
					}, rstat_string_action[string_id] == action).addClass("button_small");
				}
				
				for (i in line_by_string) {
					
					msglog.append("<hr>");
					
					var string = line_by_string[i][0].string;
					
					var event = string.event_data;
					msglog.append(event.event_str + ", " + event.match_str + ", " + event.stage_str);
					msglog.append("<br>");
					
					var show_view = true;
					var show_replace = false;
					var show_abandon = true;
					var show_resolve = false;
					
					for (j in line_by_string[i]) {
						
						var line = line_by_string[i][j];
								
						var name = "";
						var user_id = 0;
						if (string.role_shooter) user_id = string.role_shooter[line.shooter_i];
						if (user_id && g_users[user_id]) name = g_users[user_id].data.realname;
						
						var score = "";
						if (string.score && string.score[line.shooter_i]) score = string.score[line.shooter_i].rstat_score;

						if (name) msglog.append(name + " (" + score + "): ");
						
						if (line.text.indexOf("unresolved note") >= 0) {
							msglog.append("String saved with note: <b>" + string.note + "</b>");
							show_resolve = true;

						} else if (line.text.indexOf("compentry") >= 0) {
							msglog.append("Shooter not registered!");
							
						} else if (line.text.indexOf("Duplicate") >= 0) {
//Mon Sep 23 18:16:01 2019 daemon.info sh[26205]: { [Error: ER_DUP_ENTRY: Duplicate entry '118-772-4-1' for key 'PRIMARY'] code: 'ER_DUP_ENTRY', errno: 1062, sqlState: '23000', index: 0 }

							msglog.append("Score already entered!");
							show_replace = true;
								
						} else {
							msglog.append(line.text);
						}
						
						msglog.append("<br>");
						
					}
					
					if (show_view) view_button(msglog, string._id);
					if (show_replace) action_button(msglog, string._id, "Replace", "replace");
					if (show_resolve) action_button(msglog, string._id, "Resolve", "resolve");
					if (show_abandon) action_button(msglog, string._id, "Abandon", "abandon");
				}

				div.append(msglog);
			}

		} // ...if rstat
		// DMC -- a better info display?
		// shot simulator toggle
/*

		var info_div = $('<div id=infodiv>'); // viewer_shotlist.js "info" button will fill this in
		div.append(info_div);
		info_div.css("white-space", "pre"); //  so that newline and spaces are shown
		info_div.css("overflow-y", "scroll");
		info_div.css("max-height",maintarget.div.height()-$('#divToolbar').height() - $('#divTargetRow').height());
*/
//			inner.css("max-height", maintarget.div.height() * 0.5);
//		info_div.css("max-height", "10vh");
		// ..DMC

	}
	
}

toolbar_system.drawUsersEditor = function (div) {


	// select user dropdown, star range officers
	var userlist = {};
	userlist[0] = { value: 0, text: "Select a user..." };
	for (var i in g_users) {
		var user = g_users[i];
		if (user.data.name == "admin") continue;
		var str = user_display_name(i);
		if (user.data.is_range_officer) str = "(*) " + str;
		userlist[i] = { value: i, text: str };
	}

	// select user to log in as
	var item = toolbar.new_dropdown(div, userlist, function (item) {
		var id = item.val(); // val() always returns a string
		if (!id || id == "0") id = 0;

		toolbar_system.selected_user = id;
		update();
	}, toolbar_system.selected_user);

	var item = toolbar.new_button(div, "Select", function () {
		var name = prompt("Enter username / competitor number");
		if (name == null || !name || name == "0") return;

		for (var i in userlist) {
			var user_id = userlist[i].value;
			if (g_users[user_id] && name == g_users[user_id].data.name) {

				toolbar_system.selected_user = user_id;
				update();
				return;
			}
		}

		alert("username not recognized!");

	});

	toolbar.new_button(div, "New User", function () {
		var res = prompt("Enter userid for new user:");
		if (res) {
			var already_exists = false;
			for(var i in g_users) {
				if (g_users[i].data.name == res) already_exists = true;
			}
			if (already_exists) {
				alert('Userid "' + res + '" already exists')
			} else {
				var user = {};
				user._id = get_unique_id();
				user.name = res;

				var crud = { users: {} }; // build crud message
				crud.users[user._id] = user; // save new user
				g_socket.emit("crud", crud);
				toolbar_system.selected_user = user._id;
			}
		}
	});

	// for each user, delete, make range officer, etc
	var user = g_users[toolbar_system.selected_user];
	if (user && toolbar_system.selected_user) {

		// delete current user
		toolbar.new_button(div, "Delete Profile", function () {
			var res = confirm("Are you sure?");
			if (res) {
				g_socket.emit("user_delete", toolbar_system.selected_user);
				toolbar_system.selected_user = 0;
			}
		});

		toolbar.new_button(div, "Set real name", function () {
			var new_name = prompt("Enter this user's real name:");
			if (!new_name) return; // cancel
			user.crud({ realname: new_name });
		});

		if (user.data.password != '') {
			toolbar.new_button(div, "reset password", function () {
				user.crud({ password: "" });
			});
		}

		// user is promoted to range officer
		toolbar.new_checkbox(div, toolbar_military.stringname_RangeOfficer(), function () {
			user.crud({ is_range_officer: !user.data.is_range_officer });
		}, user.data.is_range_officer);
		// user is promoted to extended settings
		toolbar.new_checkbox(div, S("SYSTEM_EXTENDED_USER"), function () {
			user.crud({ is_extended_user: !user.data.is_extended_user });
		}, user.data.is_extended_user);
	}
}

toolbar_system.draw_div_backup = function(div) {
	toolbar.new_text(div, S("ADMIN_STORAGE_BACKUP"));
	toolbar.new_button(div, S("ADMIN_STORAGE_BACKUP_BUTTON"), function() {
		g_socket.emit("backup_all_data");
	});
}
toolbar_system.draw_div_restore = function(div) {
	var form = $("<form action='restore_all_data' method='post' enctype='multipart/form-data' data-ajax='false' class='title_element'>");
	form.text(S("ADMIN_STORAGE_RESTORE_ALL") + ": ");
	glob_input.change(function(e) {
		if (confirm(S("ADMIN_STORAGE_RESTORE_ALL_CONFIRM"))) {
			this.form.submit();
		}
	});
	form.append(glob_input);
	div.append(form);
}

toolbar_system.draw_div_shutdown = function(div) {
	// shutdown server
	var item = toolbar.new_button(div, S("ADMIN_SHUTDOWN"), function() {
		var answer = confirm(S("ADMIN_SHUTDOWN_CONFIRM"));
		if (!answer) return;
		g_socket.emit("shutdown");
		alert(S("ADMIN_SHUTDOWN_WARNING"));
	});
	item.css("border", "5px solid");
	item.css("font-weight", "bold");
}

// we want to aggressively trim the precision of the numbers delivered
// in a shot string.
function trim_num(nn,ndigs) {
	if (typeof ndigs == 'undefined' || ndigs < 0 || ndigs > 5) ndigs = 0;
	if (typeof nn == 'number') return +nn.toFixed(ndigs); // '+' needed to return 'number' not 'string'
	return nn;
}
