#!/bin/sh

# 
# set wifi password, for this session only, on .0.47:"SilverMountainTargets" and .0.7:"SilverMountainTargets"
# also, hide ssid .0.7:"S700"
#
# $Author: dmc $
# $Revision: 3030 $
# $Date: 2021-11-12 17:27:54 -0400 (Fri, 12 Nov 2021) $
# $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/set_wifi_password.sh $
#

S7_FSOLO_ip=192.168.0.7
alias sshQFSOLO='ssh  -y -i /www/smtv2/homedir/ssh-dir/id_rsa_smtservice.dropbear root@${S7_FSOLO_ip} 2>/dev/null'

password="$1"
encryption='none'
if [ "$password" != "" ]; then
	if [ `expr length "$password"` -ge 8 ]; then
		encryption='psk2'
	else
		echo "password length must be >= 8 chars; NOT SETTING"
		exit 1
	fi
fi

echo
echo checking server type, looking for S7 FSOLO at $S7_FSOLO_ip

# TODO: maybe do something for S25?


if wget -O - http://${S7_FSOLO_ip}/serialno_password > /dev/null 2>&1
then
	echo
	echo "we are an S7; S7 FSOLO found!  Setting S7's main processor wireless password..."
	echo
	# do for both Pi3B+ and Pi4:
	uci get  wireless.@wifi-iface[0].device > /dev/null 2>/dev/null && {
		uci set wireless.@wifi-iface[0].encryption="$encryption" 
		uci set wireless.@wifi-iface[0].key="$password"
	}

	uci get  wireless.@wifi-iface[1].device > /dev/null 2>/dev/null && {
		uci set wireless.@wifi-iface[1].encryption="$encryption" 
		uci set wireless.@wifi-iface[1].key="$password"
	}

	uci commit 2>/dev/null
	/etc/init.d/network reload
	# now un-set, so as to revert on next reboot:
	uci get  wireless.@wifi-iface[0].device > /dev/null 2>/dev/null && {
		uci set wireless.@wifi-iface[0].encryption="none" 
		uci set wireless.@wifi-iface[0].key=""
	}

	uci get  wireless.@wifi-iface[1].device > /dev/null 2>/dev/null && {
		uci set wireless.@wifi-iface[1].encryption="none" 
		uci set wireless.@wifi-iface[1].key=""
	}

	uci commit 2>/dev/null # but don't apply now; will happen on next reboot

	echo
	echo "Setting S7's wifi processor wireless password..."

	# [0] = S700, [1] = SMTserverUplink, [2] = SilverMountainTargets
	sshQFSOLO "
		uci set wireless.@wifi-iface[0].hidden=1
		uci set wireless.@wifi-iface[2].encryption=$encryption
		uci set wireless.@wifi-iface[2].key=$password
		uci commit
		/etc/init.d/network reload
		uci set wireless.@wifi-iface[0].hidden=0
		uci commit
	"

	echo
	echo FINISHED
	exit 0

else
	echo
	echo 'we are NOT an S7; FSOLO not found *** or not updated with latest firmware ***'
	exit 3
fi

exit 0

#Pi4B+
wireless.radio0=wifi-device
wireless.radio0.type='mac80211'
wireless.radio0.channel='36'
wireless.radio0.hwmode='11a'
wireless.radio0.path='platform/soc/fe300000.mmcnr/mmc_host/mmc1/mmc1:0001/mmc1:0001:1'
wireless.radio0.disabled='0'
wireless.radio0.txpower='15'
wireless.radio0.country='US'
wireless.radio0.htmode='VHT20'
wireless.default_radio0=wifi-iface
wireless.default_radio0.device='radio0'
wireless.default_radio0.network='lan'
wireless.default_radio0.mode='ap'
wireless.default_radio0.encryption='none'
wireless.default_radio0.ssid='SilverMountainTargets'
wireless.default_radio0.disabled='0'


# Pi3B+
wireless.radio0=wifi-device
wireless.radio0.type='mac80211'
wireless.radio0.path='platform/soc/3f300000.mmc/mmc_host/mmc1/mmc1:0001/mmc1:0001:1'
wireless.radio0.country='US'
wireless.radio0.legacy_rates='0'
wireless.radio0.hwmode='11a'
wireless.radio0.htmode='VHT20'
wireless.radio0.txpower='15'
wireless.radio0.channel='40'


# FSOLO
root@SMT_SOLO:/etc/config# uci show wireless
wireless.radio0=wifi-device
wireless.radio0.type='mac80211'
wireless.radio0.hwmode='11g'
wireless.radio0.path='platform/ar933x_wmac'
wireless.radio0.htmode='HT20'
wireless.radio0.channel='1'
wireless.radio0.txpower='21'
wireless.radio0.country='US'
wireless.@wifi-iface[0]=wifi-iface
wireless.@wifi-iface[0].device='radio0'
wireless.@wifi-iface[0].mode='ap'
wireless.@wifi-iface[0].encryption='none'
wireless.@wifi-iface[0].network='lan'
wireless.@wifi-iface[0].disabled='0'
wireless.@wifi-iface[0].ssid='S700'
wireless.@wifi-iface[1]=wifi-iface
wireless.@wifi-iface[1].device='radio0'
wireless.@wifi-iface[1].mode='ap'
wireless.@wifi-iface[1].key='87654321'
wireless.@wifi-iface[1].hidden='1'
wireless.@wifi-iface[1].network='lan'
wireless.@wifi-iface[1].disabled='0'
wireless.@wifi-iface[1].ssid='SMTserverUplink'
wireless.@wifi-iface[1].encryption='none'
wireless.@wifi-iface[2]=wifi-iface
wireless.@wifi-iface[2].device='radio0'
wireless.@wifi-iface[2].mode='ap'
wireless.@wifi-iface[2].encryption='none'
wireless.@wifi-iface[2].ssid='SilverMountainTargets'
wireless.@wifi-iface[2].network='lan'
wireless.@wifi-iface[2].disabled='0'
wireless.@wifi-iface[2].maxassoc='50'




